import "dart:json";
import 'media.dart';
import '../lib/pigeon.dart';
import "dart:mirrors";
const _EQ = 0;
const _GT = 1;
const _GE = 2;
const _LT = 3;
const _LE = 4;

_parseSearchCriteria(String str) {
  var parts=str.split("&&");
  var regexp=new RegExp(r"^\s*([\w_$]+)(.*)\?\s*$");
  var filters=new List(parts.length);
  int i=0;
  for (String part in parts) {
    Match match=regexp.firstMatch(part);
    if (match==null) throw "wrong expression $part";
    var columnName = match[1].trim();
    var op = match[2].trim();
    var opcode=_operations[op];
    if (opcode==null) throw "unknown operation '$op'";
    filters[i]=new Filter(columnName, opcode, i);
    i++;
  }
  return filters;  
}

// http://stackoverflow.com/questions/13306537/how-to-find-the-insertion-point-in-an-array-using-binary-search

class Filter {
  var columnName;
  var opCode;
  
  int ord;
  Filter(this.columnName, this.opCode, this.ord);
  toString() => "$columnName ${_rev_operations[opCode]} ?";
}

class IndexColumn {
  String columnName, columnType;
  int columnIndex;
  int ord;
  IndexColumn(this.columnName, this.columnType, this.columnIndex, this.ord) {
    if (columnType == null || columnIndex==null)
      throw "unknown column $columnName";
  }
}

class Index {
  var sortedRows;
  var table;
  var columns;
  Index(this.table, List columnNames) {
    columns=new List<IndexColumn>(columnNames.length);
    int i=0;
    for (var columnName in columnNames) {
      columns[i] = new IndexColumn(columnName, table.getColumnType(columnName), table.getColumnIndex(columnName), i);
      i++;
    }
  }
  rebuild() {
    var rows=table.rows;
    sortedRows=new List(rows.length);
    sortedRows.setRange(0, sortedRows.length, rows, 0);
    sortedRows.sort((a,b)=>compare(a,b));
  }
  compare(a,b) {
    for (int i=0; i<columns.length; i++) {
      var col=columns[i];
      int idx = col.columnIndex;
      var d = a.getValue(idx).compareTo(b.getValue(idx));
      if (d!=0) return d;
    }
    return 0;
  }
  compareWithValues(midVal, values) {
    for (int i=0; i<values.length; i++) {
      IndexColumn col=columns[i];
      int idx = col.columnIndex;
      var d = midVal.getValue(idx).compareTo(values[i]);
      if (d!=0) return d;
    }
    return 0;
  }
  static var _parsedCriteria = {};
  _findColumnPos(columnName) => columns.firstWhere((x) => x.columnName, ()=> throw "column not in index: $columnName").ord;
  _getFilters(searchCriteria) {
    var filters=_parsedCriteria[searchCriteria];
    var len=filters.length;
    if (filters==null) {
        filters=_parsedCriteria[searchCriteria]=_parseSearchCriteria(searchCriteria);
        filters.sort((a,b) => _findColumnPos(a.columnName) - _findColumnPos(b.columnName));
        //make sure the set of filters is compatible with binary search
        
        int nePos=filters.firstWhere((e) => e.opCode!=0, orElse: ()=>len);
        if (nePos <len-2) throw "too many inequalities";
        else if (nePos==len-2) { // special case: between
          if (filters[len-1].columnName != filters.columnName[len-2])
            throw "too many inequalities";
          int op1=filters[len-1].opCode, op2=filters[len-2].opCode;
          int pair=op1*10+op2;
          List allowed=<int>[13,31,14,41,23,32,24,42];
          if (!allowed.contains(pair))
            throw "absurd combination of conditions in the end of list";
          if (op1==_LT || op1==_LE) { // to avoid too much logic, swap them
            var temp=filters[len-1]; filters[len-1]=filters[len-2]; filters[len-2]=temp;
          }
        }
    }
    return filters;
  }
  _select(String searchCriteria, List _values) {
    var filters=_getFilters(searchCriteria);
    var len=filters.length;
    
    if (_values.length!=filters.length)
      throw "expected ${filters.length} values for criteria, got ${_values.length}";
    // filters are reshuffled, we have to reshuffle values accordingly
    // this MIGHT BE WRONG. CHECK!!!   
    var values=new List(len);
    for (int i=0; i<len; i++) values[i] = _values[filters[i].ord];

    var totalRange;
    bool extraInequality = len>=2 && filters[len-1].opCode!=_EQ && filters[len-2].opCode!=_EQ;
    if (extraInequality) {
      var effectiveValues=_cloneList(values, len-1);
      var range0= _findRange(effectiveValues);
      var values1=_cloneList(values);
      effectiveValues[len-2]=values[len-1];
      var range1=_findRange(effectiveValues);
      range0=_adjustRangeForComparisonType(range0, filters[len-2].opType);
      range1=_adjustRangeForComparisonType(range1, filters[len-1].opType);
      totalRange=_mergeRanges(range0, range1);
    } else {
      var range= _findRange(values);
      totalRange=_adjustRangeForComparisonType(range, filters[len-1].opType);
    }
    return (totalRange.length<=1) ? [] : totalRange;
    
  }
  _mergeRanges(range1, range2) {
    if (range1.length!=2 || range2.range!=2) return [];
    return [range1[0], range2[1]];
  }
  _adjustRangeForComparisonType(range, op) {
    if (range.length==1) return range;
    var hi, lo;
    switch (op) {
      case _GT: lo=range[1]+1; hi=sortedRows.length-1; break;
      case _GE: lo=range[0]; hi= sortedRows.length-1; break;
      case _LT: lo=0; hi= range[0]-1; break;
      case _LE: lo=0; hi=range[1]; break;
      case _EQ: lo=range[0]; hi=range[1]; break;
      default: throw "wrong op";
    }
    if (hi<lo) return [];
    return [lo, hi];
  }
  _cloneList(List list, [int len=-1]) {
    if (len<0) len=list.length;
    List newList=new List(len);
    for (int i=0; i<len; i++) newList[i]=list[i];
    return list;
  }
  List _binarySearchEq(List a, int low, int high, List values) {
    var lastKnownHigh = -1;
    var iter=0;
    while (low <= high) {
      if (iter++>64) throw "infinite loop";
      int mid = (low + high) >> 1;
      print("$low $high $mid");
      int midVal = a[mid];
      int diff=compareWithValues(midVal, values);
      if (diff<0) {
        low = mid + 1;
      } else if (diff>0) {
        high = mid - 1;
        lastKnownHigh=mid;
      } else if (low != mid) { //Equal but range is not fully scanned
        high = mid; //Set upper bound to current number and rescan
      } else { //Equal and full range is scanned
        return _binarySearchHigh(a, mid, lastKnownHigh, values);
      }  
    }
    return [low];  // key not found, we return insertion point
  }
  List _binarySearchHigh(List a, int low, int high, List values) {
    if (high==-1) { 
      high=a.length-1;
      var diff=compareWithValues(a[high], values);
      if (diff==0) return [low, high];
      if (diff<0) throw "bug 1";
    }
    
    var firstKnownEq=low;
    var iter=0;
    while (low <= high) {
      if (iter++>64) throw "infinite loop";
      int mid = (low + high) >> 1;
      print("HIGH: $low $high $mid");
      int midVal = a[mid];
      int diff=compareWithValues(midVal, values);
      if (diff<0)
        low = mid + 1;
      else if (diff>0)
        high = mid - 1;
      else if (high != mid) { //Equal but range is not fully scanned
        if (low==mid) return [firstKnownEq, mid];
        low = mid; //Set upper bound to current number and rescan
      }  
      else //Equal and full range is scanned
        return [firstKnownEq, mid];
    }
    throw "bug in algo";
    //return -(low + 1);  // key not found.
  }

  _findRange(values) {
    //List _binarySearchEq(List a, int low, int high, Function compare, bool knownToBeUnique) {
     return _binarySearchEq(sortedRows, 0, sortedRows.length, values);
   }
}
var _operations = {
   "==": _EQ, ">": _GT, "<": _LT, "<=": _LE, ">=": _GE                  
};
var _rev_operations=["==", ">", ">=", "<", "<="];
class Table {
  var rows=[];
  var name;
  var type;
  var metadata;
  var rowInstance;
  var indexes=[];
  bool idsAssigned=false;
  Table(this.name, this.type) {
    var mirror=reflectClass(type);
    this.rowInstance=mirror.newInstance(const Symbol(""),[]);
    this.metadata = rowInstance.reflectee.nameSet;
  }
  getColumnType(name) => metadata.getSlotType(name); 
  getColumnIndex(name) => metadata.getSlotIndex(name);
  createIndex(names) { 
    var index=new Index(this, names);
    indexes.add(index);
    return index;
  }
  
  add(PigeonMap record) => rows.add(record);
  addAll(Iterable records) => rows.addAll(records);
  assignIds() {
//    if (idsAssigned) return;
//    idsAssigned=true;
//    for (int i=0; i<rows.length; i++) rows[i].tableId_f=i;
  }
  
}
/*
 table.select("name=? && name= ?", [par1, par2]); // table can be temporary
 table.lookup("                  ", same);
 row.pseudoColumn - returns another record.
 Can they be composable? pseudoColumn.name
 Grouping: table.foldGroups(["name", "x1"], , (accum, record)=> [accum] - it should be a pigeon again,
 so we need to define pigeon for results?
 Because we return just an iterable of rows, and table is an iterable of rows
 Table is something that takes care of indexes (caches them).
 View can execute searches, too.
 Parsed requests are kept in Type. Indexes - in Table/View
 @RecordPrototype
 class Foo {
   @Computed List<Bar> get bar => barTable.select("name=? && age=?", name, age);
   @Computed int sumOfValues; // in order to avoid creating extra type? But attrs don't make sense
 }
 table.groupBy("name, age", TypeOfResult, (r) => r.sum=0, (accum, r) {}); 
 main question is how to add attributes for aggregation?  
 */
main() {
//  var table = new Table("Image",Image);
//  var filters=table.parseSearchCriteria("a!=? && a==? && b>=? && a<?");
//  print(filters);
  var x=[0,0, 1,1,1, 2,2];
  //print(_binarySearchEq(x, 0, x.length-1, 2, false));
}